﻿Imports System
Imports System.Drawing
Imports System.ComponentModel
Imports System.Windows.Forms.Layout

Namespace System.Windows.Forms.Design
	''' <summary>prevents to add more than one Control to a CaptionPanel</summary>
	Public Class CaptionPanelDesigner : Inherits ParentControlDesigner
		Protected Overrides Sub OnDragOver(e As DragEventArgs)
			If Me.Control.Controls.Count > 0 Then
				e.Effect = DragDropEffects.None
				Return
			End If
			MyBase.OnDragOver(e)
		End Sub
	End Class 'CaptionPanelDesigner
End Namespace

Namespace System.Windows.Forms

	''' <summary>
	''' Captioned Panel. Caption can be displayed on top or at left of the content. 
	''' As content only one control is allowed
	''' </summary>
	<System.ComponentModel.DesignerCategory("Code")>
	<Designer(GetType(System.Windows.Forms.Design.CaptionPanelDesigner))>
	Public Class CaptionPanel : Inherits Control : Implements ISupportInitialize

		Private _BorderStyle As BorderStyle = Windows.Forms.BorderStyle.Fixed3D

		'distinguish 2 rects: one for the caption, one for the editing-control
		Private _CaptionRect As Rectangle
		Private _ControlRect As Rectangle

		'drawing-stuff
		Private _MeasureGraphics As Graphics
		Private _TextSize As Size                                 'to cache the result of a Text-measurement
		Private _CaptionStringFormat As New StringFormat
		Private _ForeBrush As New SolidBrush(Color.Black)
		Private _BackBrush As New SolidBrush(SystemColors.Control)
		Private _ControlBackBrush As New SolidBrush(Color.AliceBlue)

		Private _CaptionOnTop As Boolean = False                 'property-backing-field

		Private _Bounds As BoundsEx
		Private _ChildBounds As BoundsEx

		Protected Overloads Overrides Sub Dispose(disposing As Boolean)
			_CaptionStringFormat.DisposeManaged(disposing).AndAlso(
				_ForeBrush, _BackBrush, _ControlBackBrush, _MeasureGraphics)
			MyBase.Dispose(disposing)
		End Sub

		Public Sub New()
			With _CaptionStringFormat
				.Alignment = StringAlignment.Center
				.LineAlignment = StringAlignment.Center
				.FormatFlags = StringFormatFlags.NoWrap
				.Trimming = StringTrimming.EllipsisCharacter
			End With
			Me.SetStyle(
				ControlStyles.ResizeRedraw Or
				ControlStyles.AllPaintingInWmPaint Or
				ControlStyles.UserPaint, True)
			TabStop = False
			_Bounds = BoundsEx.CreateEntry(Me)
			_Bounds.ReInit()
		End Sub

#Region "PublicMembers"

		<Category("Appearance")>
		<DefaultValue(BorderStyle.Fixed3D)>
		Public Property BorderStyle() As BorderStyle
			Get
				Return _BorderStyle
			End Get
			Set(value As BorderStyle)
				If _BorderStyle = value Then Return
				_BorderStyle = value
				MyBase.RecreateHandle()
				Me.PerformLayout(Me, "BorderStyle")
			End Set
		End Property

		<DefaultValue(False)>
		<Category("Appearance")>
		<Description("indicates, wether the Caption is displayed on top or left")>
		Public Property CaptionOnTop() As Boolean
			Get
				Return _CaptionOnTop
			End Get
			Set(NewValue As Boolean)
				If _CaptionOnTop = NewValue Then Return
				_CaptionOnTop = NewValue
				MyBase.Text = AddColon(MyBase.Text)
				'If Controls.Count = 0 Then SetSizeToMinimum() : Return
				Me.PerformLayout(Me, "CaptionOnTop")
			End Set
		End Property

		<TypeConverter(GetType(ExpandableObjectConverter))>
		<Browsable(True)>
		<Category("Appearance")>
		<Description("some options to vary the caption-style")>
		<DesignerSerializationVisibility(DesignerSerializationVisibility.Content)>
		Public ReadOnly Property CaptionStringFormat() As StringFormat
			Get
				If MyDesignMode.True Then Invalidate(False)
				Return _CaptionStringFormat
			End Get
		End Property

		<RefreshProperties(RefreshProperties.Repaint)>
		<Browsable(True)>
		<Category("Appearance")>
		Public Overrides Property Text() As String
			Get
				Return MyBase.Text
			End Get
			Set(value As String)
				If MyBase.Text = value Then Return
				MyBase.Text = value
				Me.PerformLayout(Me, "Text")
			End Set
		End Property

		<RefreshProperties(RefreshProperties.All)>
		<DesignOnly(True)>
		<DefaultValue(False)>
		<Description("cannot be set to true. But if you try, the" _
			& " caption re-inferres from the editing-control")>
		Public Property ReInferText() As Boolean
			Get
				Return False
			End Get
			Set(value As Boolean)
				If value Then MyBase.Text = InferCaption()
			End Set
		End Property

#Region "PublicMembers.HideFromDesigner"
		'i prefer to find only few Properties in PropertyWindow

		<Browsable(False)>
		Public Overrides Property AutoSize() As Boolean
			Get
				Return MyBase.AutoSize
			End Get
			Set(NewValue As Boolean)
				MyBase.AutoSize = NewValue
			End Set
		End Property

		<Browsable(False)>
		Public Overrides Property BackGroundImageLayout() As ImageLayout
			Get
				Return MyBase.BackgroundImageLayout
			End Get
			Set(NewValue As ImageLayout)
				MyBase.BackgroundImageLayout = NewValue
			End Set
		End Property

		<Browsable(False)>
		Public Overrides Property BackGroundImage() As Image
			Get
				Return MyBase.BackgroundImage
			End Get
			Set(NewValue As Image)
				MyBase.BackgroundImage = NewValue
			End Set
		End Property

		<Browsable(False)>
		Public Overrides Property Cursor() As System.Windows.Forms.Cursor
			Get
				Return MyBase.Cursor
			End Get
			Set(Value As System.Windows.Forms.Cursor)
				MyBase.Cursor = Value
			End Set
		End Property

		'<Browsable(False)> _
		'Public Overrides Property Anchor() As System.Windows.Forms.AnchorStyles
		'   'Anchor verbergen (SizeablePanel ist auf Docking ausgelegt)
		'   Get
		'      Return AnchorStyles.None
		'   End Get
		'   Set( Value As System.Windows.Forms.AnchorStyles)

		'   End Set
		'End Property

#End Region 'PublicMembers.HideFromDesigner
#End Region 'PublicMembers

#Region "NonPublic"

		''' <summary>calculates the required CaptionPanel-Size from a childControls Size</summary>
		Private Function GetSizeFromChild(childSize As Size) As Vector
			Dim szMe = _TextSize + Padding.Size
			If CaptionOnTop Then
				szMe.Height = (_TextSize.Height * 3) \ 2 + childSize.Height
				szMe.Width.Maximize(childSize.Width + Padding.Horizontal)
			Else
				szMe.Width += childSize.Width
				szMe.Height.Maximize(childSize.Height + Padding.Vertical)
			End If
			GetSizeFromChild = szMe + (Size - ClientSize) 'add Border-Size
			If Controls.Count = 0 Then GetSizeFromChild.Y.Maximize(20)
		End Function

		''' <summary>get me new Bounds and Minimumsize</summary>
		Private Sub UpdateBoundsEx(minimize As Boolean)
			With _Bounds
				.Min = GetSizeFromChild(SizeControlMin)
				If _ChildBounds.Null Then
					.Size = .Min
					.CanSize.ClearWith(True)
				Else
					Dim sz = GetSizeFromChild(_ChildBounds.Ctl.Size)
					.Size = If(minimize, sz, sz.Union(.Size))
					Array.Copy(_ChildBounds.CanSize, .CanSize, 2) 'infer my CanSize from Child
				End If
				Me.MinimumSize.Assign(.Min.ToSize)
				Me.Size.Assign(.Size.ToSize)
			End With
		End Sub

		''' <summary>
		''' calculate and set _CaptionRect and _ControlRect from the actual CaptionPanel-Size 
		''' </summary>
		Private Sub SetRects()
			_CaptionRect = ClientRectangle
			Dim amount As Integer
			With ClientRectangle
				If _CaptionOnTop Then
					amount = (_TextSize.Height * 3) \ 2
					_CaptionRect.Height = amount
					_ControlRect = New Rectangle(.X, amount, .Width, .Height - amount)
				Else
					amount = _TextSize.Width
					_CaptionRect.Width = amount
					_ControlRect = New Rectangle(amount, .Y, .Width - amount, .Height)
				End If
			End With
		End Sub
		Private Shared _rgxPrefix As New Text.RegularExpressions.Regex("^[a-z]+")
		''' <summary>examines the childControl and proposes a caption</summary>
		Private Function InferCaption() As String
			If Me.Controls.Count = 0 Then Return AddColon(Name)
			With Controls(0)
				If .DataBindings.Count = 0 Then
					Dim s = .GetType.Name
					If .Name.EndsWith(s) Then Return AddColon(.Name.Substring(0, .Name.Length - s.Length))
					If s.EndsWith("DataGridView") Then
						With TryCast(DirectCast(.Self, DataGridView).DataSource, BindingSource)
							If .NotNull Then
								Dim tb = .DataTable
								If tb.NotNull Then Return AddColon(.DataTable.TableName)
							End If
						End With
					End If
					Return AddColon(_rgxPrefix.Replace(.Name, ""))
				Else
					Dim bnd = .DataBindings(0)
					Return AddColon(bnd.BindingMemberInfo.BindingMember)
				End If
			End With
		End Function

		'append/remove ':', depending wether the Caption is on the left side or on Top
		Private Function AddColon(s As String) As String
			s = s.TrimEnd(":"c, " "c)
			Return If(_CaptionOnTop, s, s & ":")
		End Function

#End Region 'NonPublic

#Region "ProcessingEvents"

		Protected Overrides ReadOnly Property CreateParams() As CreateParams
			Get 'For Borderstyle there is to recreate Params
				Dim cp As CreateParams = MyBase.CreateParams
				Select Case _BorderStyle
					Case BorderStyle.FixedSingle
						cp.Style = (cp.Style Or &H800000)
					Case BorderStyle.Fixed3D
						cp.ExStyle = (cp.ExStyle Or &H200)
				End Select
				Return cp
			End Get
		End Property

		Protected Overrides Sub OnTextChanged(e As EventArgs)
			MyBase.OnTextChanged(e)
			ReLayout()
		End Sub
		Protected Overrides Sub OnFontChanged(e As EventArgs)
			MyBase.OnFontChanged(e)
			ReLayout()
		End Sub
		Private Sub ReLayout()
			If _MeasureGraphics.Null Then _MeasureGraphics = Me.CreateGraphics
			_TextSize = _MeasureGraphics.MeasureString(Text, Font, SizeBig, _CaptionStringFormat).Ceiling
			Invalidate(_CaptionRect)
		End Sub

		Protected Overrides Sub OnBackColorChanged(e As EventArgs)
			MyBase.OnBackColorChanged(e)
			_BackBrush.Color = BackColor
		End Sub
		Protected Overrides Sub OnForeColorChanged(e As EventArgs)
			MyBase.OnForeColorChanged(e)
			_ForeBrush.Color = ForeColor
		End Sub
		Protected Overrides Sub OnPaint(e As System.Windows.Forms.PaintEventArgs)
			'draw control-Background, caption-Background and caption
			If Controls.Count = 1 Then
				Dim col = Controls(0).BackColor
				If _ControlBackBrush.Color.ToArgb <> col.ToArgb Then _ControlBackBrush.Color = col
				e.Graphics.FillRectangle(_ControlBackBrush, _ControlRect)
			End If
			e.Graphics.FillRectangle(_BackBrush, _CaptionRect)
			e.Graphics.DrawString(MyBase.Text, Font, _ForeBrush, _CaptionRect, _CaptionStringFormat)
		End Sub

		Protected Overrides Sub OnMouseDown(e As System.Windows.Forms.MouseEventArgs)
			'forward the Focus to the editing-Control
			MyBase.OnMouseDown(e)
			If Controls.Count = 1 Then Controls(0).Focus()
		End Sub

		Protected Overrides Sub OnControlAdded(e As System.Windows.Forms.ControlEventArgs)
			If Controls.Count > 1 Then
				Dim msg = String.Format("{0} - (""{1}""): can't take more than 1 Control!", Me.Name, Me.Text)
				If MyDesignMode.True Then MessageBox.Show(msg)
				Throw New ArgumentException(msg)
			End If
			MyBase.OnControlAdded(e)
		End Sub

		Protected Overrides Sub OnLayout(e As LayoutEventArgs)
			If Parent.Null Then Return
			If _Bounds.Null Then  'occurs in Designmode with currently added CaptionPanel. On the other hand:
				'°                       EndInit() is called at runtime or when the Form is opened in Designer
				_Bounds = Me.BoundsX
				If Controls.Count > 0 Then _ChildBounds = Controls(0).BoundsX
			End If
			Parent.SuspendLayout()
			If e.AffectedControl Is Me Then
				Select Case e.AffectedProperty
					Case "Padding", "CaptionOnTop", "BorderStyle"
						UpdateBoundsEx(True)
					Case "Bounds" 'prevent sizing if CanSize(i)=False
						With _Bounds
							Dim vNew As Vector = Me.Size
							For i = 0 To 1
								If .CanSize(i) Then .Size(i) = vNew(i)
							Next
							Me.Size.Assign(.Size.ToSize)
						End With
				End Select
			Else
				Select Case e.AffectedProperty
					Case "Parent" 'a childControl is added or removed in Designer
						If Controls.Count > 0 Then
							_ChildBounds = Controls(0).BoundsX
						Else
							_ChildBounds = Nothing
							UpdateBoundsEx(False)
						End If
					Case "Bounds", "AutoSize", "Font"
						_ChildBounds.AutoDetectCanSize()
						UpdateBoundsEx(True)
				End Select
			End If
			SetRects()
			If Controls.Count > 0 Then
				Dim rct = _ControlRect.Subtract(Padding)
				'set ChildControl to rct. If sizing is limited, center it in rct.
				With _ChildBounds
					.Size = rct.Size
					.Location = .ApplyLimitation.Scale(0.5) + rct.Location
					.ApplyBounds(False)
					.ReInit()
				End With
			End If
			'MsgDsn("cp", "-" & e.AffectedControl.Name & "-", e.AffectedProperty, old, Size.ToVector)
			_Bounds.ReInit()
			Parent.ResumeLayout(True)
			MyBase.OnLayout(e)
		End Sub
		Public Overrides ReadOnly Property LayoutEngine() As Layout.LayoutEngine
			Get
				Return NoLayout.Instance
			End Get
		End Property

		Public Sub BeginInit() Implements ISupportInitialize.BeginInit
		End Sub

		Public Sub EndInit() Implements ISupportInitialize.EndInit
			_Bounds = Me.BoundsX
			If Controls.Count > 0 Then _ChildBounds = Controls(0).BoundsX
			UpdateBoundsEx(False)
			SetRects()
		End Sub

#End Region 'ProcessingEvents
	End Class
End Namespace